;-------------------------------------------;
;                 Real Mode                 ;
;        Multitasking/multithreading        ;
;                   Demo                    ;
;                                           ;
;       by Alexei A. Frounze (c) 2000       ;
;                                           ;
; E-mail:   alexfru@chat.ru                 ;
; Homepage: http://alexfru.chat.ru          ;
; Mirror:   http://members.xoom.com/alexfru ;
;-------------------------------------------;

IDEAL
MODEL   LARGE
P386

STACK_SIZE      EQU     4096
MAX_TASKS       EQU     8
FASTER          EQU     8

EXTRN   OLD_TIMER   : DWORD
EXTRN   SWITCHING   : BYTE
EXTRN   BLOCKS      : DWORD
EXTRN   STACK_LIST  : DWORD
EXTRN   TASK_CNT    : WORD
EXTRN   CUR_TASK    : WORD
EXTRN   AVAIL_TIME  : DWORD
EXTRN   ACTIVE_TIME : DWORD

PUBLIC  SHEDULER, ADD_TASK

        CODESEG

PROC    SHEDULER FAR
        PUSH    DS
        PUSH    ES
        PUSH    FS
        PUSH    GS
        PUSH    EAX
        PUSH    EBX
        PUSH    ECX
        PUSH    EDX
        PUSH    ESI
        PUSH    EDI
        PUSH    EBP                     ; SAVE TASK REGISTERS ON THE STACK

        MOV     AX, DGROUP
        MOV     DS, AX                  ; LOAD DS WITH VALID DATA SEGMENT

        INC     [TICK]
        CMP     [TICK], FASTER
        JE      @@CALL
        MOV     AL, 20H
        OUT     20H, AL
        JMP     @@SKIP

@@CALL:
        MOV     [TICK], 0
        PUSHF
        CALL    [OLD_TIMER]             ; HANDLE IRQ0 AT STANDARD FRQUENCY

@@SKIP:
        CMP     [SWITCHING], 0
        JE      @@EXIT                  ; SWITCHING IS DISABLED

        CMP     [TASK_CNT], 2
        JC      @@EXIT                  ; NOTHING TO SWITCH

        MOV     AX, [CUR_TASK]
        MOV     BX, AX
        SHL     BX, 2

        LEA     SI, [AVAIL_TIME]
        MOV     CX, [DS:BX+SI]          ; CX = AVAILABLE TIME
        LEA     SI, [ACTIVE_TIME]
        INC     [WORD PTR DS:BX+SI]
        CMP     [DS:BX+SI], CX
        JC      @@EXIT                  ; SWITCHING WILL BE A BIT LATER
        MOV     [WORD PTR DS:BX+SI], 0

        LEA     SI, [STACK_LIST]        ; DS:SI = ^TASK LIST (STACK LIST)
        MOV     [DS:BX+SI], SP
        MOV     [DS:BX+SI+2], SS        ; SAVE CURRENT STACK STATE

@@SEARCH:
        INC     AX                      ; NEXT TASK NUMBER
        CMP     AX, [TASK_CNT]
        JC      @@NEXT
        XOR     AX, AX                  ; FIRST TASK
@@NEXT:
        MOV     [CUR_TASK], AX          ; NEXT TASK NUMBER

        MOV     BX, AX
        SHL     BX, 2

        LEA     SI, [AVAIL_TIME]
        CMP     [WORD PTR DS:BX+SI], 0
        JE      @@SEARCH                ; KEEP SEARCHING, IF FROZEN TASK FOUND

        LEA     SI, [STACK_LIST]        ; DS:SI = ^TASK LIST (STACK LIST)
        MOV     SP, [DS:BX+SI]
        MOV     SS, [DS:BX+SI+2]        ; RESTORE STACK STATE

@@EXIT:
        POP     EBP
        POP     EDI
        POP     ESI
        POP     EDX
        POP     ECX
        POP     EBX
        POP     EAX
        POP     GS
        POP     FS
        POP     ES
        POP     DS                      ; RESTORE TASK REGISTERS FROM THE STACK
        IRET                            ; CONTINUE THE TASK
ENDP

PROC    ADD_TASK PASCAL FAR     _IP:WORD, _CS:WORD, _DS:WORD, _AT:WORD
        CMP     [TASK_CNT], MAX_TASKS
        JAE     @@EXIT

        LEA     SI, [BLOCKS]            ; DS:SI = ^BLOCKS
        MOV     BX, [TASK_CNT]
        SHL     BX, 2
        MOV     AX, [DS:BX+SI]
        MOV     DX, [DS:BX+SI+2]        ; DX:AX = ^BLOCK

        ADD     AX, STACK_SIZE-42       ; CORRECT FOR SP

        LEA     SI, [STACK_LIST]        ; DS:SI = ^STACK PTR
        MOV     [DS:BX+SI], AX
        MOV     [DS:BX+SI+2], DX        ; SS:SP IS DONE FOR THE TASK

        PUSH    ES
        MOV     ES, DX
        ADD     AX, 42
        MOV     DI, AX                  ; ES:DI = ^STACK

        MOV     [WORD PTR ES:DI-2], 202H; FLAGS
        MOV     AX, [_CS]
        MOV     [ES:DI- 4], AX          ; CS
        MOV     AX, [_IP]
        MOV     [ES:DI- 6], AX          ; IP
        MOV     AX, [_DS]
        MOV     [ES:DI- 8], AX          ; DS
        MOV     [ES:DI-10], AX          ; ES
        XOR     EAX, EAX
        MOV     [ES:DI-12], AX          ; FS
        MOV     [ES:DI-14], AX          ; GS
        MOV     [ES:DI-18], EAX         ; EAX
        MOV     [ES:DI-22], EAX         ; EBX
        MOV     [ES:DI-26], EAX         ; ECX
        MOV     [ES:DI-30], EAX         ; EDX
        MOV     [ES:DI-34], EAX         ; ESI
        MOV     [ES:DI-38], EAX         ; EDI
        MOV     [ES:DI-42], EAX         ; EBP

        POP     ES

        LEA     SI, [AVAIL_TIME]
        MOV     AX, [_AT]
        MOV     [DS:BX+SI], AX

        LEA     SI, [ACTIVE_TIME]
        MOV     [WORD PTR DS:BX+SI], 0

        INC     [TASK_CNT]

@@EXIT:
        RET
ENDP

        DATASEG

TICK    DW     0

END
